/*
 *  particle.cpp
 *  smoke
 *
 *  Created by CBreak on Wed May 04 2005.
 *  Copyright (c) 2005 __MyCompanyName__. All rights reserved.
 *
 */

#include "particle.h"

#include <iostream.h>
#include <math.h>
#include "types.h"

const int gridsize=4;

// Particle Colour, in RGB Tripples, one sec temporal delta
const int colourCount = 3*7;
const float particleColour[colourCount] = {1,0,0, 1,1,0, 1,1,1, 0,1,1, 0,0,1, 0,0,0, 0,0,0};

// Management
Particle::Particle()
{
	// Initialize (x,y,z) in [-16..16]^3
	s.x = (float)rand()/RAND_MAX*32-16; s.y = (float)rand()/RAND_MAX*32-16; s.z = (float)rand()/RAND_MAX*32-16;
	commonInit();
	inventorInit();
};

Particle::Particle(float x, float y, float z)
{
	s.x = x; s.y = y; s.z = z;
	commonInit();
	inventorInit();
};

Particle::Particle(float x, float y, float z, SoCamera * cam)
{
	s.x = x; s.y = y; s.z = z;
	commonInit();
	inventorInit();
//	myCam = cam;
	particleTransform->rotation.appendConnection(&(cam->orientation));
};

inline void Particle::commonInit()
{
	age = 0;
	v.x = 0; v.y = 0; v.z = 0;
	a.x = 0; a.y = 0; a.z = 0;	
}

inline void Particle::inventorInit()
{
	static SoTriangleStripSet * particleVisual = 0;
//	static SoPointSet * particleVisual = 0;
	
	if (particleVisual == 0) // Only create once
	{
		// Initialize static shapes
		particleVisual = new SoTriangleStripSet;
		particleVisual->numVertices = 4;
//		particleVisual = new SoPointSet;
		particleVisual->ref(); // Never meant to be released
	}
	particleTransform = new SoTransform;
	particleTransform->ref();
	particleTransform->translation.setValue(s.x, s.y, s.z);
	particleMaterial = new SoMaterial;
	particleMaterial->ref();
	particleMaterial->diffuseColor.setHSVValue(0,0,0);
	particleMaterial->emissiveColor.setHSVValue(0,1,1);
	this->addChild(particleTransform);
	this->addChild(particleMaterial);
	this->addChild(particleVisual);
}

Particle::~Particle()
{
	this->removeChild(particleTransform);
	particleTransform->unref();
	this->removeChild(particleMaterial);
	particleMaterial->unref();
};

// Animation
void Particle::move(float t)
{
//	a.x -= v.x*0.01; a.y -= v.y*0.01; a.z -= v.z*0.01; // damping
// Semi implicite leapfrog euler type of scheme...
	v.x += a.x*t; v.y += a.y*t; v.z += a.z*t;
	s.x += v.x*t; s.y += v.y*t; s.z += v.z*t;
	particleTransform->translation.setValue(s.x, s.y, s.z);
	a.x = 0; a.y = 0; a.z = 0; // reset accelerations for next pass
	
//	particleTransform->rotation = myCam->orientation; // Orientate like camera
	// problem: Only executed with active movement.
	
	age += t;
/*  float ageTemp=age/6;
	if (ageTemp>1.0) ageTemp=1;
	particleMaterial->emissiveColor.setHSVValue(ageTemp, 1, 1);*/
	float colourRGB[3];
	float ageFrac = age - floor(age);
	colourRGB[0] =  (1-ageFrac)*particleColour[(3*(int)floor(age))%colourCount] +
					ageFrac*particleColour[(3*(int)ceil(age))%colourCount];
	colourRGB[1] =  (1-ageFrac)*particleColour[(3*(int)floor(age)+1)%colourCount] +
		ageFrac*particleColour[(3*(int)ceil(age)+1)%colourCount];
	colourRGB[2] =  (1-ageFrac)*particleColour[(3*(int)floor(age)+2)%colourCount] +
		ageFrac*particleColour[(3*(int)ceil(age)+2)%colourCount];
	particleMaterial->emissiveColor.setValue(colourRGB);

	//particleMaterial->emissiveColor.setHSVValue(age/6, 1, 1);
	particleMaterial->transparency = age/6.2;
}

void Particle::adjustAcceleration(float fx, float fy, float fz)
{
	//float m = 1; // Mass asumed to be 1
	a.x += fx; a.y += fy; a.z += fz;
}

void Particle::adjustAcceleration(vec3f f)
{
	//float m = 1; // Mass asumed to be 1
	a.x += f.x; a.y += f.y; a.z += f.z;
}

void Particle::interactWith(Particle * p)
{
	float f, d; // force, distance
	vec3f r = p->getPosition();
	r.x -= s.x; r.y -= s.y; r.z -= s.z; // direction
	d = sqrt(r.x*r.x + r.y*r.y + r.z*r.z);
	if (d<2*gridsize) // cutoff distance
	{
		int c=1;
		// Use Lennard Jones force
		float t1 = d*d, t2 = t1*t1;
		if(d>0.8)
			f = -c*(1/t2 - 1.2/t1)/d; // We could cap the applied force
		else
			f = -c*(0.708+(1-d));
		//f = -(1/(pow(t2,1)) - 1/pow(t1,1))*c/d; // Slower due pow
		adjustAcceleration(f*r.x, f*r.y, f*r.z);
	}
}

// Accessors
vec3f Particle::getPosition()
{
	return s;
}

// Debug
void Particle::print()
{
	cout << "(" << s.x << "," << s.y << "," << s.z << ")\n";
}
