/*
 *  world.cpp
 *  smoke
 *
 *  Created by CBreak on Sat May 07 2005.
 *  Copyright (c) 2005 __MyCompanyName__. All rights reserved.
 *
 */

#include "world.h"

#include <iterator>
#include <iostream>

#include "types.h"
#include "platform.h"

#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoMaterialBinding.h>
//#include <Inventor/draggers/SoTrackballDragger.h>
//#include <Inventor/manips/SoTrackballManip.h>
//#include <Inventor/manips/SoJackManip.h>
#include <Inventor/nodes/SoRotation.h>


using namespace std;

static float stdFrameInterval = 0.04; // 25 fps

static float floorCorodinates[][3] = { -32,-4,-32, -32,-4,32, 32,-4,-32, 32,-4,32 };
static float particleCorodinates[][3] = { -2,-2,0, 2,-2,0, -2,2,0, 2,2,0 };


// for interaction of particles
const int arraySize=257; // Prime Number
vector<Particle*> *  particleArray[arraySize];

World::World()
{	
	root = new SoSeparator();
	particles = new vector<Particle*>();
	
	vector<Particle*> * particleVect;
	for (int i=0; i<arraySize; i++)
	{
		particleVect = new vector<Particle*>();
		particleArray[i]= particleVect;
    }

	// Camera setup
	cam = new SoPerspectiveCamera();
	cam->position = SbVec3f(0, 16, 64);
	cam->nearDistance = 1;
	cam->farDistance = 128;
	cam->pointAt(SbVec3f(0,4,0));
	root->addChild(cam);

	// Emitter
	theEmitter = new DragEmitter(this);
	//theEmitter= new MovingEmitter(this);
	root->addChild(theEmitter);
	
	// Model Space
//	SoTrackballManip * myDragger = new SoTrackballManip;
//	SoJackManip * myDragger = new SoJackManip;
//	SoTrackballDragger * myDragger = new SoTrackballDragger;
//	root->addChild(myDragger);
	part = new SoTransformSeparator;
	root->addChild(part);
//	SoRotation * myRotation = new SoRotation;
//	myRotation->rotation.appendConnection(&(myDragger->rotation));
//	part->addChild(myRotation);
	
	// Floor
	SoBaseColor * color = new SoBaseColor;
	color->rgb.setValue(1, 1, 1);
	part->addChild(color);

	SoCoordinate3 * rectangleCoords = new SoCoordinate3;
	rectangleCoords->point.setValues(0, 4, floorCorodinates);
	part->addChild(rectangleCoords);

	SoTriangleStripSet * rectangleTriangle = new SoTriangleStripSet;
	rectangleTriangle->numVertices = 4;
	part->addChild(rectangleTriangle);
	
	// Set up stuff for the particles
	rectangleCoords = new SoCoordinate3;
	rectangleCoords->point.setValues(0, 4, particleCorodinates);
	part->addChild(rectangleCoords);
	SoTexture2 * particleTexture = new SoTexture2;
	particleTexture->filename.setValue(getPathForItem(PPParticleTexture)); // Get Platform Path
	part->addChild(particleTexture);

	SoMaterialBinding * materialBinding = new SoMaterialBinding();
	materialBinding->value = SoMaterialBinding::PER_PART;
	part->addChild(materialBinding);
	
	// Animation Timer Setup. Could be an Idle Callback too
	tSensor = new SoTimerSensor(idleCallback, this);
	tSensor->setInterval(stdFrameInterval);
	tSensor->schedule();
	
	// Force Setup
	theForce = new VirtualForcefield();
	
	lastMove = getCurrentTime();

	state = running;
	interact = false;
};

World::~World()
{
	delete particles;
	delete theForce;
};

// Accessors
SoSeparator * World::getRoot()
{
	return root;
}

SoCamera * World::getCam()
{
	return cam;
}


void World::addParticle(Particle * someThing)
{
	particles->push_back(someThing); // And to mathematical representation
	part->addChild(someThing); // Add to graphical representation
	particleArray[hash(someThing->getPosition())]->push_back(someThing); 

}

int World::hash (int i, int j, int k)
{	
	int hashValue;
	int gridLength=4;
	
    hashValue= abs((23*(i/gridLength)+43*(j/gridLength)+73*(k/gridLength))%arraySize);
	
	return hashValue;
}

int World::hash (vec3f v)
{	
	return hash((int)v.x, (int)v.y,(int)v.z);
}


// Management
void World::resetParticles()
{
	vector<Particle*>::iterator a = particles->begin();
	vector<Particle*>::iterator z = particles->end();
	
	// Remove old particles
	while (a != z)
	{
		part->removeChild(*a); // Implicite dealocation
		a++;
	}
	particles->clear();
}

void World::resetForces()
{
	delete theForce;
	theForce = new VirtualForcefield;
}

// Emitter
void World::setEmissionRate(float newRate)
{
	theEmitter->setRate(newRate);
}

// Animation
void World::move()
{
	thisMove = getCurrentTime();
	float frameInterval = (float)(thisMove-lastMove); // probably not precise enough
	lastMove = thisMove;
	
	vector<Particle*>::iterator a;
	vector<Particle*>::iterator b;
	vector<Particle*>::iterator z;

	if (interact)
	{
		a = particles->begin();
		z = particles->end();
		while (a != z)
		{
			vec3f pos = (*a)->getPosition();
			for(int j = (int)(pos.x-1); j<=pos.x+1; j++){
				for(int k = (int)(pos.y-1); k<=pos.y+1; k++){
					for(int l = (int)(pos.z-1); l<= pos.z+1; l++)
					{
						vector<Particle*>::iterator i1 = particleArray[hash(j,k,l)]->begin();
						vector<Particle*>::iterator i2 = particleArray[hash(j,k,l)]->end();

						while (i1 != i2)
						{		 
							if (*a != *i1)
							(*a)->interactWith(*i1);
							
							i1++;
						}

					}
				}
			}
			a++;
		}
	}

	a = particles->begin();
	z = particles->end();
	
	while (a != z)
	{
		// Clean, but inefficient I guess...
		(*a)->adjustAcceleration(theForce->getForceAtPoint((*a)->getPosition()));
		a++;
	}
    
	for (int i=0; i<arraySize; i++)
	{	
		particleArray[i]->clear();
    }
	
	a = particles->begin();
	b = particles->begin();
	z = particles->end();

	while (b != z)
	{
		if ((*b)->age > 6) // Die after 6 secs
		{
			part->removeChild(*b); // Implicite dealocation
			b++;
		}
		else
			break;
	}
		
	particles->erase(a, b);
	
	a = particles->begin();
	z = particles->end();
	while (a != z)
	{
		(*a)->move(frameInterval);
	    particleArray[hash((*a)->getPosition())]->push_back(*a);
		a++;
	}

	theEmitter->emitt(frameInterval);
}

void World::toggle()
{
	switch(state)
	{
		case running:
			pause();
			break;
		case paused:
			run();
			break;
	}
}

void World::pause()
{
	tSensor->unschedule();
	state = paused;	
}

void World::run()
{
	tSensor->schedule();

	lastMove = getCurrentTime();

	state = running;
}

void World::addTForce()
{
	theForce->attachField(new TForce(0.5));
}

void World::addGForce()
{
	theForce->attachField(new GForce(3));
}

void World::toggleInteraction()
{
	if(interact)
		interact = false;
	else
		interact = true;
}


// Callbacks
void idleCallback(void * data, SoSensor * sensor)
{
	((World *)data)->move();
//	sensor->schedule(); // Neccessary with SoIdleSensor
}
